char version[12] = "2015-05-07";

/*  Copyright (C) 2014-15 R. D. Tennent School of Computing,
 *  Queen's University, rdt@cs.queensu.ca
 *
 *  This program is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU General
 *  Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will
 *  be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A
 *  PARTICULAR PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General
 *  Public License along with this program; if not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street,
 *  Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*  fixmsxpart - corrects note spacing in a single-staff 
 *               musixtex part
 */

# include "utils.h"

# define NNOTES "\\vnotes1.6\\elemskip"
# define NNNOTES "\\vnotes1.28\\elemskip"
# define Notespp "\\vnotes3.75\\elemskip"
# define NOtespp "\\vnotes4.85\\elemskip"
# define APPOGG_NOTES  "\\vnotes0.85\\elemskip"
# define APPOGG_SPACING  96  /* three 128ths  */

# define SHORT_LEN 256

PRIVATE char infilename[SHORT_LEN];
PRIVATE char *infilename_n = infilename;
PRIVATE char outfilename[SHORT_LEN];
PRIVATE char *outfilename_n = outfilename;
PRIVATE char logfilename[SHORT_LEN];
PRIVATE char *logfilename_n = logfilename;

PRIVATE FILE *infile, *outfile, *logfile;

PRIVATE int lineno = 0;

PRIVATE bool autospaced = false;
PRIVATE char line[LINE_LEN];    /* line of input                                        */
PRIVATE int staff = 1;          /* staff number                                         */
PRIVATE char *notes;            /* notes or other material for staff part               */
PRIVATE bool no_notes = true;
/* used to prevent outputting a line feed when a notes command has generated no output  */

PRIVATE char *current;
PRIVATE int spacing = 0; /* 0 = not determined                                          */
  /* 1 = whole or breve, 2 = half, 4 = quarter, 8 = eigth, etc.                         */
PRIVATE int beaming = 0;        /* beam-note spacing                                    */
PRIVATE int new_beaming = 0;    /* spacing to be used *after* the next beam note        */
PRIVATE int restbars = 0;       /* number of whole-bar rests not yet output             */
PRIVATE char deferred_bar[SHORT_LEN];

PRIVATE int cspacing;
PRIVATE char collective[SHORT_LEN];
PRIVATE bool first_collective;

PRIVATE void
usage (FILE *f)
{
  fprintf (f, "Usage: fixmsxpart [-v | --version | -h | --help]\n");
  fprintf (f, "       fixmsxpart [-a | --autospaced] infile[.tex] [outfile[.tex]] \n");
}

PRIVATE void
error_at (const char msg[])      /* abort with stderr message msg                        */
{
  fprintf (stderr, "Error on line %d: %s\n", lineno, msg);
  exit (EXIT_FAILURE);
}

PRIVATE void
status (void)
{ 
  if (notes == NULL)
    fprintf (logfile, "notes=NULL\n");
  else
    fprintf (logfile, "notes=%s\n", notes); 
  if (current == NULL)
    fprintf (logfile, "current=NULL\n");
  else
    fprintf (logfile, "current=%s\n", current); 
  fprintf (logfile, "spacing=%d\n", spacing);
  fprintf (logfile, "restbars=%d\n", restbars);
  fflush (logfile);
}

PRIVATE void
status_beam (void)
{
  fprintf (logfile, "beaming=%d\n", beaming);
  fprintf (logfile, "new_beaming=%d\n", new_beaming);
  fprintf (logfile, "spacing=%d\n", spacing);
  fflush (logfile);
}

PRIVATE void
status_collective (void)
{
  fprintf (logfile, "cspacing=%d\n", cspacing);
  fprintf (logfile, "collective=%s\n", collective);
  fprintf (logfile, "first_collective=%d\n", first_collective);
  fflush (logfile);
}

PRIVATE char 
*analyze_notes (char **ln) {
/*
   skip initial TeX command, which may be unreliable; 
   then skip staff-1 ampersands or vertical bars; 
   then truncate tail and return a pointer to the 
   beginning of the staff commands.
   ln is moved to the token after \en
*/
  int i; char *s; char *t; char *tt;
  s = strpbrk (*ln+1, "|&\\\n"); /* skip initial command      */
  while (true)
  { /* look for \en */
    t = strstr(s, "\\en");
    if (t != NULL) break;
    else
    /* replace EOL by a blank and append another line of input */
    { char new_line[LINE_LEN];
      if (fgets (new_line, LINE_LEN, infile) == NULL)
        error_at("Unexpected EOF.");
      if ((t = strchr (s, '\n')) == NULL )
        error_at("Missing EOL.");
      *t = ' ';
      t++;
      *t = '\0';
      if (append (line, &t, new_line, LINE_LEN) >= LINE_LEN)
        error_at ("Line too Long.");
    }
  }
  for (i=1; i<staff; i++)      /* skip staff-1 & or |       */
  { 
     s = strpbrk (s, "|&");
     if (s == NULL) error_at ("Can't find & or |.");
     s++;
  }
  t = strstr(s, "\\en"); 
  if (t == NULL) error_at ("Can't find \\en.");
  *ln = t+3;
  tt = strpbrk (s, "|&");
  if ( tt != NULL && tt < t) t = tt;
  *t = '$'; *(t+1) = '\0';
  return s;
}


PRIVATE void
output_notes (char *t) 
{ /* outputs text from *notes up to *t (except for 
     commands \sk \hsk \noteskip= \multnoteskip);
     then updates notes.
  */
  char *s;
    fprintf (logfile, "\nBefore output_notes:\n");
    status ();
    fprintf (logfile, "\noutputting:");
  s = notes;
  if (collective[0] == '\0' || first_collective)
  { while (s < t) 
    { char *tt;
      /* discard  commands \sk, \hsk, \qsk, \noteskip= ..., \multnoteskip  */ 
      if ( prefix("\\sk", s)  ||
           prefix("\\hsk", s)  ||
           prefix("\\qsk", s)  ||
           prefix("\\noteskip", s) ||
           prefix("\\multnoteskip", s ) )
      { 
        s = strpbrk (s+1, "\\$");
        if (s >= t) break;
        continue;
      }
      tt = strpbrk (s+1, "\\$");
      if (tt >= t) tt = t;
      while (s < tt)
      { putc (*s, outfile); putc (*s, logfile); s++; }
 
    }
  }
  if (collective[0])
  { 
    if (!first_collective)
    {
      fprintf (outfile, "%s", collective);
      fprintf (logfile, "%s", collective);
    }
    first_collective = false;
    while (s < t)
    { if (*s != '*' || cspacing > 16 ) 
      {
        fputc (*s, outfile); fputc (*s, logfile);
      }
      s++; 
    }
    fprintf (outfile, "%s", "}");
    fprintf (logfile, "%s", "}");
    if (*t == '}')
    { collective[0] = '\0';
      t++;
    }
      fprintf (logfile, "\n");
      fprintf (logfile, "\nAfter outputting collective note:\n");
      status ();
      status_collective ();
      notes = t; current = notes;
      return;
  }
  notes = t; current = notes;
    fprintf (logfile, "\n");
    fprintf (logfile, "\nAfter output_notes\n");
    status ();
  return;
}

PRIVATE void
output_rests (void)
{ /* outputs a multi-bar rest and the deferred_bar command */
  switch (restbars)
  { case 0: return; 
    case 1: 
      fprintf ( outfile, "\\NOTEs\\sk\\en\n" );
      fprintf ( outfile, "\\def\\atnextbar{\\znotes\\centerpause\\en}%%\n" );
      break;
    case 2: 
      fprintf ( outfile, "\\NOTesp\\sk\\en\n" );
      fprintf ( outfile, "\\def\\atnextbar{\\znotes\\centerbar{\\zcharnote{9}{\\meterfont\\lrlap2}}\\centerbar{\\PAuse}\\en}%%\n" );
      break;
    case 3:
      fprintf ( outfile, "\\NOTes\\hqsk\\sk\\en\n" );
      fprintf ( outfile, "\\def\\atnextbar{\\znotes\\centerbar{\\zcharnote{9}{\\meterfont\\lrlap3}}\\centerbar{\\PAuse\\off{2.5\\elemskip}\\pause}\\en}%%\n" );
      break;
    case 4:
      fprintf ( outfile, "\\NOtesp\\sk\\en\n" );
      fprintf ( outfile, "\\def\\atnextbar{\\znotes\\centerbar{\\zcharnote{9}{\\meterfont\\lrlap4}}\\centerbar{\\PAUSe}\\en}%%\n" );
      break;
    case 5:
      fprintf ( outfile, "\\NOTes\\hqsk\\sk\\en\n" );
      fprintf ( outfile, "\\def\\atnextbar{\\znotes\\centerbar{\\zcharnote{9}{\\meterfont\\lrlap5}}\\centerbar{\\PAUSe\\off{2.5\\elemskip}\\pause}\\en}%%\n" );
      break;
    case 6:
      fprintf ( outfile, "\\NOTes\\hqsk\\sk\\en\n" );
      fprintf ( outfile, "\\def\\atnextbar{\\znotes\\centerbar{\\zcharnote{9}{\\meterfont\\lrlap6}}\\centerbar{\\PAUSe\\off{2.5\\elemskip}\\PAuse}\\en}%%\n" );
      break;
    case 7:
      fprintf ( outfile, "\\NOTEs\\hqsk\\sk\\en\n" );
      fprintf ( outfile, "\\def\\atnextbar{\\znotes\\centerbar{\\zcharnote{9}{\\meterfont\\lrlap7}}\\centerbar{\\PAUSe\\off{2.5\\elemskip}\\PAuse\\off{2.5\\elemskip}\\pause}\\en}%%\n" );
      break;
    case 8:
      fprintf ( outfile, "\\NOTes\\hqsk\\sk\\en\n" );
      fprintf ( outfile, "\\def\\atnextbar{\\znotes\\centerbar{\\zcharnote{9}{\\meterfont\\lrlap8}}\\centerbar{\\PAUSe\\off{2.5\\elemskip}\\PAUSe}\\en}%%\n" );
      break;
    case 9:
      fprintf ( outfile, "\\NOTEs\\hqsk\\sk\\en\n" );
      fprintf ( outfile, "\\def\\atnextbar{\\znotes\\centerbar{\\zcharnote{9}{\\meterfont\\lrlap9}}\\centerbar{\\PAUSe\\off{2.5\\elemskip}\\PAUSe\\off{2.5\\elemskip}\\pause}\\en}%%\n" );
      break;
    default:  
      fprintf ( outfile, "\\NOTEs\\hqsk\\Hpause4{0.83}\\en\n" );
      fprintf ( outfile, "\\def\\atnextbar{\\znotes\\centerbar{\\zcharnote{9}{\\meterfont\\lrlap{%d}}}\\en}%%\n", restbars );
  }
  if (restbars > 1) 
    fprintf ( outfile, "\\advance\\barno%d%%\n", restbars-1 );
  restbars = 0; 
  fprintf ( outfile, "%s\n", deferred_bar); 
}

PRIVATE void
initialize_notes (void) /* start a new notes command  */
{ switch (spacing) 
  { case 128:fprintf (outfile, "%s", NNNOTES); break;
    case 96: fprintf (outfile, "%s", APPOGG_NOTES); break;
    case 64: fprintf (outfile, "%s", NNNOTES); break;
    case 32: fprintf (outfile, "%s", NNOTES); break;
    case 24: fprintf (outfile, "\\notesp"); break;
    case 16: fprintf (outfile, "\\notes"); break;
    case 14: fprintf (outfile, "%s", Notespp); break;
    case 12: fprintf (outfile, "\\Notesp"); break;
    case 8:  fprintf (outfile, "\\Notes"); break;
    case 7:  fprintf (outfile, "%s", NOtespp); break;
    case 6:  fprintf (outfile, "\\NOtesp"); break;
    case 4:  fprintf (outfile, "\\NOtes"); break;
    case 3:  fprintf (outfile, "\\NOTesp"); break;
    case 2:  fprintf (outfile, "\\NOTes"); break;
    case 1:  fprintf (outfile, "\\NOTEs"); break;
    case 0:  fprintf (outfile, "\\znotes"); break;
    default: 
      fprintf (stderr, "Spacing value %d not recognized.\n", spacing);
      exit (EXIT_FAILURE);
  }
  fprintf (logfile, "\noutputting notes command for spacing=%d\n", spacing);
}


PRIVATE void semiautomatic_beam (char *s)
{
  if ( prefix ("\\Dqbb", s) 
    || prefix ("\\Tqbb", s)
    || prefix ("\\Qqbb", s) )
    beaming = 16; 
  else 
  if ( prefix ("\\Dqb", s)
    || prefix ("\\Tqb", s)
    || prefix ("\\Qqb", s) )
  { beaming = 8; }
  else
    error ("Can't recognize semiautomatic beam command");
  if (spacing > 0) fprintf (outfile, "\\en\n");
  output_rests ();
  if (beaming == 16) 
    fprintf (outfile, "\\notes"); 
  else /* beaming == 8 */
    fprintf (outfile, "\\Notes");
  current = strpbrk (s+1, "\\$");
  output_notes (current);
  spacing = beaming;
  no_notes = false;
}


PRIVATE
int collective_note (void)
{ char *s = current; char *t;
  int spacing = cspacing;
  t = strchr (s, '}');
    fprintf (logfile, "\nEntering collective_note:\n");
    status ();
    status_collective ();
    fprintf (logfile, "t=%s\n", t);
  while (s < t) /* search for alphabetic note */
  { if (*s == '.' && new_beaming == 0 && autospaced) 
      spacing = spacing * 1.50; 
    else if (isalpha (*s) )
    { s++; 
      current = s;
        fprintf (logfile, "\nAfter collective_note:\n");
        status ();
        status_collective ();
      return spacing; 
    }
    s++;
  }
  current = s++;
    fprintf (logfile, "\nAfter collective_note:\n");
    status ();
    status_collective ();
  return spacing;
}


PRIVATE void beam_initiation (char *s) 
{ /* sets beaming, but doesn't output anything yet */
  int new_beaming = 8;
  if ( prefix ("\\ibbbbbb", s)
    || prefix ("\\Ibbbbbb", s)
    || prefix ("\\nbbbbbb", s) )
    new_beaming = 256; 
  else if ( prefix ("\\ibbbbb", s)
    || prefix ("\\Ibbbbb", s)
    || prefix ("\\nbbbbb", s) )
    new_beaming = 128; 
  else if ( prefix ("\\ibbbb", s)
    || prefix ("\\Ibbbb", s)
    || prefix ("\\nbbbb", s) )
    new_beaming = 64; 
  else if ( prefix ("\\ibbb", s) 
    || prefix ("\\Ibbb", s)
    || prefix ("\\nbbb", s) ) 
    new_beaming = 32; 
  else if ( prefix ("\\ibb", s) 
    || prefix ("\\Ibb", s)
    || prefix ("\\nbb", s) )
    new_beaming = 16; 
  else if ( prefix ("\\ib", s) 
    || prefix ("\\Ib", s) )
    new_beaming = 8; 
  if (new_beaming > beaming)  
    beaming = new_beaming;
}

PRIVATE void beam_termination (char *s)  /* \tb... */
{ 
  if ( prefix ("\\tbbbbbb", s) )
  { new_beaming = 128; if (beaming < 256) beaming = 256; }
  else if ( prefix ("\\tbbbbb", s) )
  { new_beaming = 64; if (beaming < 128) beaming = 128; }
  else if ( prefix ("\\tbbbb", s) )
  { new_beaming = 32; if (beaming < 64) beaming = 64; }
  else if ( prefix ("\\tbbb", s) )
  { new_beaming = 16; if (beaming < 32) beaming = 32; }
  else if ( prefix ("\\tbb", s) ) 
  { new_beaming = 8; if (beaming < 16) beaming = 16; }
  else if ( prefix ("\\tb", s) )
  { new_beaming = 0; beaming = 8; } 
  return;
}

PRIVATE void nonspacing_beam_termination (void)   /*  \ztq...  */
{ 
  new_beaming = 0; beaming =8;
  return;
}

PRIVATE int spacing_note (void)
{ 
  char *s, *t;
  bool dotted = false;
  bool doubledotted = false;
  int new_spacing = 0;
  if (collective[0])
  { 
    new_spacing = collective_note (); 
    return new_spacing;
  }

  s = current;
  if (*s == '$') return 0;
  while (true)
  {
      fprintf (logfile, "\nIn spacing_note: s=%s\n", s);
      fflush (logfile);
    if ( prefix ("\\hs", s) 
       && !prefix ("\\hsk", s)  
       && !prefix ("\\hsp", s) ) 
    { new_spacing = 32; break;}

    if ( prefix ("\\hsp", s ) )
    { new_spacing = 48; break; }

    if ( prefix ("\\h", s) 
       && !prefix ("\\hqsk", s)  
       && !prefix ("\\hsk", s)  
       && !prefix ("\\hpp", s) ) 
    { new_spacing = 2; break;}

    if ( prefix ("\\hpp", s ) )
    { new_spacing = 3; break; }

    if ( prefix ("\\qs", s)
       && !prefix ("\\qsk", s) 
       && !prefix ("\\qsp", s) )
    { new_spacing = 16; break; }

    if ( prefix ("\\qsp", s ) )
    { new_spacing = 24; break; }

    if ( prefix ("\\q", s)
       && !prefix ("\\qb", s) 
       && !prefix ("\\qsk", s) 
       && !prefix ("\\qqsk", s)  
       && !prefix ("\\qqs", s)  
       && !prefix ("\\qpp", s) ) 
    { new_spacing = 4; break; }

    if ( prefix ("\\qpp", s ) )
    { new_spacing = 6; break; }

    if ( prefix ("\\qqs", s )
       && !prefix ("\\qqsp", s) 
       && !prefix ("\\qqsk", s) )
    { new_spacing = 64; break; }

    if ( prefix ("\\qqsp", s) )
    { new_spacing = 96; break; }

    if ( prefix ("\\cccc", s) ) 
    { new_spacing = 64; break; }

    if ( prefix ("\\ccc", s) ) 
    { new_spacing = 32; break; }

    if ( prefix ("\\cc", s) && !prefix("\\cchar", s) )
    { new_spacing = 16; break; }

    if ( prefix ("\\ds", s ) 
       && !prefix ("\\dsp", s)  
       && !prefix ("\\dsh", s) ) 
    { new_spacing = 8; break; }

    if ( prefix ("\\dsp", s ) )
    { new_spacing = 12; break; }

    if ( prefix ("\\c", s) 
       && !prefix ("\\cna", s)  
       && !prefix ("\\csh", s)  
       && !prefix ("\\cfl", s)  
       && !prefix ("\\curve", s)  
       && !prefix ("\\ccharnote", s) )
    { new_spacing = 8; break; }

    if ( prefix ("\\wh", s)
      || prefix ("\\breve", s) )
    { new_spacing = 1; break; }

    if ( prefix ("\\qb", s) ) /* beam note */
    { 
        fprintf (logfile, "\nBeam note:\n");
        status_beam ();
      output_rests ();
      new_spacing = beaming;  
      if (new_beaming != 0)  /* set by preceding \tb... */
      { beaming = new_beaming; new_beaming = 0; }
      break; 
    }

    if ( prefix("\\tq", s) ) /* beam completion */
    { new_spacing = beaming; 
      new_beaming = 0; 
      beaming = 8;
      break; 
    }


  /* not a spacing note:  */

    if ( prefix("\\ib",s) 
       || prefix("\\Ib",s) 
       || prefix("\\nb",s) ) 
      beam_initiation (s); 

    else if ( prefix("\\tinynotesize", s)) 
    { int tmp_spacing;
      current = strchr (s+1, '\\');
      tmp_spacing = spacing_note ();
      return APPOGG_SPACING;
    }

    else if ( prefix("\\tb", s) ) 
      beam_termination (s);

    else if ( prefix("\\ztq", s) ) 
      nonspacing_beam_termination (); 

    else if ( prefix("\\Dq",s) 
       || prefix("\\Tq",s) 
       || prefix("\\Qq",s) )
    {  semiautomatic_beam (s); 
         fprintf (logfile, "\nAfter semiautomatic beam:\n");
         status ();
       return beaming;
    }

    else if ( prefix("\\zchar", s) 
       || prefix("\\lchar", s)
       || prefix("\\cchar", s) 
       || prefix("\\zcn", s)
       || prefix("\\lcn", s)
       || prefix("\\ccn", s) )
    { /* skip two arguments (second may have embedded commands) */
      s = strchr (s+1, '}'); /* first }  */
      s = strchr (s+1, '}'); /* second } */
    }
   
    else if (prefix("\\ppt", s) 
       || prefix("\\pppt", s) )
      doubledotted = true;  /* triple-dotted spaced as double-dotted */
      
    else if (prefix("\\pt", s)) 
      dotted = true;
      
    /* non-spacing; skip to next command */
    s = strpbrk(s+1,"\\$"); 
    if (*s == '$') 
    { current = s; return new_spacing; }

  } /* end of while (true) */

  no_notes = false;

  if (prefix ("\\hupp", s)
   || prefix ("\\hlpp", s)
   || prefix ("\\qupp", s)
   || prefix ("\\qlpp", s)
   || prefix ("\\cupp", s)
   || prefix ("\\clpp", s)
   || prefix ("\\qbpp", s) 
   || doubledotted  )
  {
    new_spacing *= 1.75; doubledotted = false;
  }

  else if (prefix ("\\whp", s)
   || prefix ("\\hup", s)
   || prefix ("\\hlp", s)
   || prefix ("\\qup", s)
   || prefix ("\\qlp", s)
   || prefix ("\\cup", s)
   || prefix ("\\clp", s)
   || prefix ("\\qbp", s) 
   || dotted  )
  {
    new_spacing *= 1.5; dotted = false;
  }

  t = strpbrk (s+1, "{\\$"); /* collective coding?  */
  if (*t == '{')  /*  {...}  */
  { /* Save prefix in collective.
     * It will be output for every note in the collective */
    char *ss = s; 
    char *tt = collective;
    while (ss <= t)
    { *tt = *ss; tt++, ss++; }
    *tt = '\0';
    current = t+1;
    cspacing = new_spacing;
    first_collective = true;
    return collective_note ();
  } 

  current = strpbrk (s+1, "\\$");
  return new_spacing;
}

PRIVATE 
bool skips (void)
/* to avoid outputting \znotes\en */
{ char *s = notes;
  while (*s != '$')
  {
    if (!prefix ("\\sk", s)) return false;
    s = strpbrk (s+1, "\\$");
  }
  notes = s;
  return true;
}

PRIVATE void
synthesize_notes (void)
/* 
  
   Generate well-spaced notes .

   A notes segment is of the form
  
        ...\XXX...\XXX...\XXX...$
        ^      ^
      notes   current

   where ... is non-spacing material (possibly empty) and
   \XXX is a spacing command. The terminating $ was appended
   in analyze_notes.

   synthesize_notes reads through non-spacing material
   (updating current) until a spacing command is found.

   If the spacing for the spacing command is the same as the
   currently active spacing, the material from notes up to
   current is output, and the notes pointer is updated.

   If the spacing for the spacing command differs, a new
   notes command is initialized, after possibly terminating
   a currently active notes command.

   Then repeat until *current is $.

*/
{ 
  if (skips()) { spacing = 0; return; }
  current = notes;
  do
  { int new_spacing;
    new_spacing = spacing_note ();
      fprintf (logfile, "\nAfter spacing_note:\n");
      fprintf (logfile, "new_spacing=%d\n", new_spacing);
      status ();
    if (new_spacing != spacing || new_spacing == 0 )
    { 
      if (spacing != 0) 
      { 
        fprintf (outfile, "\\en"); fprintf (logfile, "\n\noutputting:\\en\n"); 
        if (skips()) { spacing = 0; return; }
        else putc ('\n', outfile); 
      }
      spacing = new_spacing;
      if (spacing > 0 && spacing < APPOGG_SPACING) output_rests ();
        fprintf (logfile, "\nBefore initialize_notes:\n");
        status ();
      initialize_notes ();
    }
    output_notes (current);
  }
  while (*current != '$');
  fprintf (outfile, "\\en"); fprintf (logfile, "\n\noutputting:\\en\n");
  spacing = 0;
  return;
}

PRIVATE void
process_command (char **ln)
{ char *s, *t;

  if ( prefix("\\def\\vnotes#1\\elemskip", *ln) ) 
  { /* determine staff */
    s = strstr (*ln, "@vnotes") + 7;
    t = strchr(s, '#'); /* may have commands before # */
    if (sscanf (t, "#%d", &staff) != 1) error_at("sscanf for staff fails");
    staff--; 
    fprintf (outfile,"\\def\\vnotes#1\\elemskip#2\\en{\\noteskip#1\\@l@mskip\\@vnotes");
    while (s != t)  
    { putc(*s, outfile); s++; }
    fprintf (outfile,"#2\\enotes}");
    *ln = strchr (*ln, '}') ;
    (*ln)++;
  }

  else if ( prefix("\\TransformNotes", *ln) ) 
  { /* determine staff */
    s = strchr (*ln, '}');
    if (s == NULL) error_at ("Can't parse \\TransformNotes.");
    s = strchr (s+1, '{');
    if (s == NULL) error_at ("Can't parse \\TransformNotes.");
    s++;
    t = strchr(s, '#'); /* may have \transpose... before # */
    if (sscanf (t, "#%d", &staff) != 1) error_at("sscanf for staff fails");
    staff--; 
    fprintf (outfile,"\\TransformNotes{#2}{");
    while (s != t)
    { putc(*s, outfile); s++; }
    fprintf (outfile,"#2}");
    *ln = strchr (t, '}') ;
    if (*ln == NULL) error_at ("Can't find '}' after \\TransformNotes.\n");
    (*ln)++;
  }

  else if ( prefix("\\nnnotes", *ln) ||  
            prefix("\\nnotes", *ln) ||
            prefix("\\notes", *ln) || 
            prefix("\\Notes", *ln) ||
            prefix("\\NOtes", *ln) ||
            prefix("\\NOTes", *ln) ||
            prefix("\\NOTEs", *ln) ||
            prefix("\\NOTES", *ln) )
  { /*  spaced notes */
      fprintf (logfile, "\nBefore analyze_notes:\n");
      fprintf (logfile, "*ln=%s\n", *ln);
      fflush (logfile);
    notes = analyze_notes (ln);
      fprintf (logfile, "\nAfter analyze_notes:\n");
      status ();
    spacing = 0;
    no_notes = true;
    current = notes;
        fprintf (logfile, "\nBefore synthesize_notes:\n");
      status ();
    synthesize_notes ();
      fprintf (logfile, "\nAfter synthesize_notes:\n");
      fprintf (logfile, "*ln=%s\n", *ln);
      status ();
    t = strpbrk (*ln, "%\\\n");
    *ln = t;
    if (no_notes) 
      (*ln)++;  /* prevents outputting a blank line   */
  }

  else if ( prefix("\\vnotes", *ln) )
  { 
    char *s; char *t;
    s = *ln+7;
    t = strchr ( s, '\\');
    *ln = t; /* skip to \elemskip */
    notes = analyze_notes (ln);
    if ( strstr (notes, "\\tinynotesize") != NULL )
    { /* appoggiatura note  */
      fprintf (logfile, "\nAppoggiatura note\n");
      fprintf (logfile, "outputting:");
      output_rests ();
      fprintf (outfile,"\\vnotes");
      fprintf (logfile,"\\vnotes");
      while (s!=t) 
      { putc(*s, outfile); s++; } 
      t = strchr (notes, '$'); 
      *t = '\0';
      fprintf (outfile,"\\elemskip%s\\en", notes);
      fprintf (logfile,"\\elemskip%s\\en\n", notes);
      t = strpbrk (*ln, "%\\\n");
      if (t == NULL) t = *ln + strlen (*ln);
      *ln = t;
    }
    else
    { 
        fprintf (logfile, "\nBefore synthesize_notes:\n");
        status ();
      spacing = 0;
      no_notes = true;
      synthesize_notes ();
      t = strpbrk (*ln, "%\\\n");
      *ln = t;
      if (no_notes) 
        (*ln)++;  /* avoids blank line if no notes  */
    }
  }

  else if ( prefix("\\znotes", *ln) )
  { /*  non-spacing material */
    output_rests ();
    notes = analyze_notes (ln);
    t = strchr (notes, '$');
    *t = '\0';
    fprintf (outfile,"\\znotes%s\\en", notes);
    t = strpbrk (*ln, "%\\\n");
    if (t == NULL) t = *ln + strlen (*ln);
    *ln = t;
  }

  else if ( prefix("\\def\\atnextbar{\\znotes", *ln))
  { /*  whole-bar or multi-bar rest? */
    *ln = *ln + 15;
    notes = analyze_notes(ln);
    if (strstr (notes, "\\centerpause") != NULL )
    { restbars++;   /* defer output until after a   */
                    /* possible multi-bar rest      */
    }
    else
    {
      if (*notes != '$') 
      { fprintf (outfile,"\\def\\atnextbar{\\znotes%s\\en}", notes);}
    }
    t = strchr (*ln, '}');
    if (t == NULL) error_at ("Can't find }.");
    *ln = t+1;
    t = strchr (*ln, '\\');
    if (t == NULL ) t = *ln + strlen(*ln);
    *ln = t;
  }

  else if ( (prefix("\\bar", *ln) && !prefix("\\barno", *ln) ) 
         || prefix("\\doublebar", *ln)
         || prefix("\\ala", *ln) )
       { 
         if ( prefix("\\doublebar", *ln) )
         { if (restbars > 0)
             sprintf (deferred_bar, "\\doublebar");
           else
             fprintf (outfile, "\\doublebar");
         }
         else if ( prefix("\\alaligne", *ln) )
         { if (restbars > 0)
             sprintf (deferred_bar, "\\alaligne");
           else
             fprintf (outfile, "\\alaligne");
         }
         else if ( prefix("\\alapage", *ln) )
         {
           if (restbars > 0)
             sprintf (deferred_bar, "\\alapage");
           else
             fprintf (outfile, "\\alapage");
         }
         else
         { if (restbars > 0)
             sprintf (deferred_bar, "\\bar");
           else
             fprintf (outfile, "\\bar");
         }
         t = strpbrk (*ln+1, "%\\");
         if (t == NULL) t = *ln + strlen (*ln); 
         *ln = t;
       }

  else if ( prefix("\\Endpiece", *ln) )
  { if (restbars > 0)
    {
      sprintf (deferred_bar, "%s", *ln);
      output_rests ();
    }
    else
      fprintf (outfile,"%s", *ln);
    *ln = *ln + strlen (*ln);
  }

  else if ( prefix("\\endpiece", *ln) )
  { if (restbars > 0)
    {
      sprintf (deferred_bar, "%s", *ln);
      output_rests ();
    }
    else
      fprintf (outfile,"%s", *ln);
    *ln = *ln + strlen (*ln);
  }

  else if ( prefix("\\instrumentnumber", *ln) )
  { fprintf (outfile, "\\instrumentnumber1");
    t = strpbrk (*ln+1, "%\\");
    if (t == NULL) 
    { *ln = *ln + strlen (*ln); fprintf (outfile, "\n"); return; }
    *ln = t;
  }

  else if ( prefix("\\startpiece", *ln) )
  { fprintf (outfile, "%s", "\\setstaffs11\n\\nostartrule\n\\startpiece");
    t = strpbrk (*ln+1, "%\\\n");
    if (t == NULL) t = *ln + strlen (*ln);
    *ln = t;
  }

  else if ( prefix("\\mulooseness", *ln) 
         || prefix("\\linegoal", *ln)  
         || prefix("\\song", *ln)
         || prefix("\\group", *ln) 
         || prefix("\\akkoladen", *ln) )
  { fputc ('%', outfile);
    t = strpbrk (*ln+1, "%\\\n");
    if (t == NULL) 
    { fprintf (outfile, "%s", *ln); return; }
    while (*ln < t) { fputc (**ln, outfile); (*ln)++; } 
    fputc ('\n', outfile);
  }

  else if ( prefix("\\def", *ln) )
  { char msg[LINE_LEN];
    sprintf (msg, "Macro definition ignored: %s", *ln);
    fprintf (stderr, "Warning: %s", msg);
    t = *ln + strlen(*ln);
    while (*ln < t)
    { fputc (**ln, outfile); 
      (*ln)++;
    }
  }

  else  /* everything else */
  { t = strpbrk (*ln+1, "%\\");
    if (t == NULL || *t == '%') t = *ln + strlen (*ln);
    while (*ln < t) 
    { fputc (**ln, outfile); 
      (*ln)++;
    }
  }
}

PRIVATE void
process_line (void)
{
  char  *ln; 
  ln = &line[0];
  while ( *ln != '\0') 
  {
    while (*ln == ' ') { ln++; fputc (' ', outfile); }
    if (*ln == '%') {
      fprintf (outfile, "%s", ln);
      return;
    }
    process_command (&ln);  
  }     /* *ln == '\0'  */
}

PRIVATE void
process_score (void)
/* process .tex file */
{
  int c;
  while ( (c = getc (infile)) != EOF )
  {
    ungetc (c, infile);
    lineno++;
    if ( fgets(line, LINE_LEN, infile) == NULL) error_at("IO error");
    if (strlen (line) == LINE_LEN-1) error_at("Line too long.");
    fprintf (stderr, "\nLine %d:%s", lineno, line);
    process_line (); 
  }     /* c == EOF  */
}     


int 
main (int argc, char *argv[])
{
  int c;
  char today[12];
  time_t mytime; 
# define NOPTS 4
  struct option longopts[NOPTS] =
  {  { "help", 0, NULL, 'h'},
     { "version", 0, NULL, 'v'},
     { "autospaced", 0, NULL, 'a'},
     { NULL, 0, NULL, 0}
  };
  
  time (&mytime);
  strftime (today, 11, "%Y-%m-%d", localtime (&mytime) );
  c = getopt_long (argc, argv, "hva", longopts, NULL);
  while (c != -1)
    {
      switch (c)
        {
        case 'h':
          fprintf (stdout, "This is fixmsxpart, version %s.\n", version);
          usage (stdout);
          fprintf (stdout, "Please report bugs to rdt@cs.queensu.ca.\n" );
          exit (0);
        case 'v':
          fprintf (stdout, "This is fixmsxpart, version %s.\n", version);
          exit (0);
        case 'a':
          autospaced = true; break;
        case '?':
          exit (EXIT_FAILURE);
        default:
          fprintf (stderr,"Function getopt returned character code 0%o.\n",
                  (unsigned int) c);
          exit (EXIT_FAILURE);
        }
      c = getopt_long (argc, argv, "hva", longopts, NULL);
    }
  fprintf (stderr, "This is fixmsxpart, version %s.\n", version);
  fprintf (stderr, "Copyright (C) 2014-15  R. D. Tennent\n" );
  fprintf (stderr, "School of Computing, Queen's University, rdt@cs.queensu.ca\n" );
  fprintf (stderr, "License GNU GPL version 2 or later <http://gnu.org/licences/gpl.html>.\n" );
  fprintf (stderr, "There is NO WARRANTY, to the extent permitted by law.\n\n" );

  infilename[0] = '\0';
  infilename_n = infilename;
  if ( (optind < argc) && (optind+2 >= argc))
    {
      append (infilename, &infilename_n, argv[optind], sizeof (infilename));
      if (!suffix (".tex", infilename))
      append (infilename, &infilename_n, ".tex", sizeof (infilename));
    }
  else 
  {  usage (stderr);
     exit (EXIT_FAILURE);
  }
  optind++;
  outfilename[0] = '\0';
  outfilename_n = outfilename;
  if (optind < argc)  /* user-provided outfilename */
    {
      append (outfilename, &outfilename_n, argv[optind], sizeof (outfilename));
      if (!suffix (".tex", outfilename))
        append (outfilename, &outfilename_n, ".tex", sizeof (outfilename));
    }
  infile = fopen (infilename, "r");
  if (infile == NULL)
    {
      fprintf (stderr,"Can't open %s\n", infilename);
      exit (EXIT_FAILURE);
    }
  fprintf (stderr, "Reading from %s.", infilename);
  if (*outfilename == '\0')
  {
    outfile = stdout;
    fprintf (stderr, " Writing to stdout.");
  }
  else
  {
    outfile = fopen (outfilename, "w");
    if (outfile == NULL)
    {
      fprintf (stderr,"Can't open %s\n", outfilename);
      exit (EXIT_FAILURE);
    }
    fprintf (stderr, " Writing to %s.", outfilename);
  }

  logfilename[0] = '\0';
  logfilename_n = logfilename;
  append (logfilename, &logfilename_n, infilename, sizeof (logfilename));
  logfilename_n -= 4;
  *logfilename_n = '\0';
  append (logfilename, &logfilename_n, ".flog", sizeof (logfilename));
  logfile = fopen (logfilename, "w");
  if (logfile == NULL)
    {
      fprintf (stderr, "Can't open %s\n", logfilename);
      exit (EXIT_FAILURE);
    }
  fprintf (stderr, " Log file %s.", logfilename);
  fprintf (stderr,"\n\n");


  fprintf (outfile, "%% Generated on %s by fixmsxpart (%s).\n", today, version);
  fprintf (logfile, "Generated on %s by fixmsxpart (%s).\n", today, version);
  process_score (); 

  return 0;
}
